/* $Id  */

/*
 *      Author: 	Jeff Hill  
 *      Date:          	5-95 
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *              The Controls and Automation Group (AT-8)
 *              Ground Test Accelerator
 *              Accelerator Technology Division
 *              Los Alamos National Laboratory
 *
 *      Co-developed with
 *              The Controls and Computing Group
 *              Accelerator Systems Division
 *              Advanced Photon Source
 *              Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00  mm-dd-yy        iii     Comment
 */

#ifndef INCepicsTypesh
#define INCepicsTypesh 1

#ifdef __STDC__
#	define READONLY const
#else
#	define READONLY 
#endif

#include <shareLib.h>
#include <stdio.h>

#ifndef stringOf
#	ifdef __STDC__ 
#		define stringOf(TOKEN) #TOKEN
#	else
#		define stringOf(TOKEN) "TOKEN"
#	endif
#endif

typedef enum {
	epicsFalse=0, 
	epicsTrue=1 } 	epicsBoolean;

/*
 * Architecture Independent Data Types
 * (so far this is sufficient for all archs we have ported to)
 */
typedef char		epicsInt8;
typedef unsigned char   epicsUInt8;
typedef short           epicsInt16;
typedef unsigned short  epicsUInt16;
typedef epicsUInt16	epicsEnum16;
typedef int             epicsInt32;
typedef unsigned 	epicsUInt32;
typedef float           epicsFloat32;
typedef double          epicsFloat64;
typedef struct {
	unsigned	length;
	char		*pString;
}epicsString;
typedef	unsigned long	epicsIndex;
typedef struct {
	epicsIndex	first;
	epicsIndex	count;
}epicsArrayIndex;

/*
 * To avoid redundant copying of large buffers
 * the buffer provider is required to supply a
 * function pointer (pFree) which is called when
 * consumer is finished with the buffer.
 * The provider must _not_ change or free the
 * buffer until the pFree call back is called.
 *
 * To avoid redundant copying of large buffers
 * the buffer provider may break the buffer up into 
 * many non-contiguous segments each of which is
 * pointed to (and chained together) by an epicsSegment.
 */
typedef struct epics_segment {
	void			*pBuf;
	struct epics_segment	*pNext;
	void			(*pFree)(void *pBuf, void *pArg);
	void			*pArg;
	unsigned long		bufSize; /* bytes */
}epicsSeg;
#ifdef __STDC__
typedef void epicsSegFree (void *pBuff, void *pArg);
typedef void *epicsSegMalloc (size_t size, void *pArg);
#else
typedef void epicsSegFree ();
typedef void *epicsSegMalloc ();
#endif

/*
 *
 * ndim - the number of dimensions 
 * 		ndim==0 => scaler
 *		otherwise pLengths points to an array of ndim items
 * pLengths - pointer to an array of lengths for
 *		each dimension in the array
 */
typedef struct {
	unsigned 	ndim;	
	epicsIndex	*pLengths;
	epicsSeg	*pSeg;
}epicsArray;

/*
 * !! Dont use this - it may vanish in the future !!
 *
 * Provided only for backwards compatibility with
 * db_access.h 
 *
 */
#define MAX_STRING_SIZE 40
typedef char            epicsOldString[40]; 

/*
 * union of all types
 *
 * Strings included here as pointers only so that we support
 * large string types.
 *
 * Arrays included here as pointers because large arrays will 
 * not fit in this union.
 */
typedef union epics_any{
        epicsInt8       int8;
        epicsUInt8      uInt8;
        epicsInt16      int16;
        epicsUInt16     uInt16;
        epicsEnum16     enum16;
        epicsInt32      int32;
        epicsUInt32     uInt32;
        epicsFloat32    float32;
        epicsFloat64    float64;
	epicsString	string;
	epicsArray	*pArray;
}epicsAny;
	
/*
 * IN: 		nDim - the number of dimensions 
 *		pMalloc - pointer to a function that is to allocate this struct
 *				(if NULL then malloc() is called)
 *		pFree 	- pointer to a function that will be called to free this 
 *				structure
 *				(if NULL then free() is called)
 *		... - list of sizes for each dimension starting with dimension zero
 * RETURN: 	pointer to an epicsArray structure
 *
 * EXAMPLE:
 * epicsCreateArray(0, NULL, NULL, NULL) - scaler array 	
 * epicsCreateArray(1, NULL, NULL, NULL, 10) - one dimensional array
 * epicsCreateArray(2, NULL, NULL, NULL, 12, 140) - two dimensional array
 */
epicsArray *epicsCreateArray(unsigned nDim, epicsSegMalloc *pMalloc, 
		epicsSegFree *pFree, ...);

/*
 * Calls any segment free functions that may exist
 * for an EPICS data items that use segments 
 */
void epicsDeleteArray(epicsArray *pData);

/*
 * Corresponding Type Codes
 * (this enum must start at zero)
 *
 * !! Update epicsTypeToDBR_XXXX[] and DBR_XXXXToEpicsType
 * 	in db_access.h if you edit this enum !!
 */
typedef enum {
                epicsInt8T,
                epicsUInt8T,
                epicsInt16T,
                epicsUInt16T,
                epicsEnum16T,
                epicsInt32T,
                epicsUInt32T,
                epicsFloat32T,
                epicsFloat64T,
                epicsStringT,
                epicsOldStringT
}epicsType;
#define firstEpicsType epicsInt8T
#define lastEpicsType epicsOldStringT 
#define validEpicsType(x) 	((x>=firstEpicsType) && (x<=lastEpicsType))
#define invalidEpicsType(x)	((x<firstEpicsType) || (x>lastEpicsType))


/*
 * The enumeration "epicsType" is an index to this array
 * of type name strings.
 */
#ifdef epicsTypesGLOBAL
READONLY char *epicsTypeNames [lastEpicsType+1] = {
		stringOf (epicsInt8),
		stringOf (epicsUInt8),
		stringOf (epicsInt16),
		stringOf (epicsUInt16),
		stringOf (epicsEnum16),
		stringOf (epicsInt32),
		stringOf (epicsUInt32),
		stringOf (epicsFloat32),
		stringOf (epicsFloat64),
		stringOf (epicsString),
		stringOf (epicsOldString),
};
#else /* epicsTypesGLOBAL */
epicsShareExtern READONLY char *epicsTypeNames [lastEpicsType+1];
#endif /* epicsTypesGLOBAL */

/*
 * The enumeration "epicsType" is an index to this array
 * of type code name strings.
 */
#ifdef epicsTypesGLOBAL
READONLY char *epicsTypeCodeNames [lastEpicsType+1] = {
		stringOf (epicsInt8T),
		stringOf (epicsUInt8T),
		stringOf (epicsInt16T),
		stringOf (epicsUInt16T),
		stringOf (epicsEnum16T),
		stringOf (epicsInt32T),
		stringOf (epicsUInt32T),
		stringOf (epicsFloat32T),
		stringOf (epicsFloat64T),
		stringOf (epicsStringT),
		stringOf (epicsOldStringT),
};
#else /* epicsTypesGLOBAL */
epicsShareExtern READONLY char *epicsTypeCodeNames [lastEpicsType+1];
#endif /* epicsTypesGLOBAL */

#ifdef epicsTypesGLOBAL
READONLY unsigned epicsTypeSizes [lastEpicsType+1] = {
		sizeof (epicsInt8),
		sizeof (epicsUInt8),
		sizeof (epicsInt16),
		sizeof (epicsUInt16),
		sizeof (epicsEnum16),
		sizeof (epicsInt32),
		sizeof (epicsUInt32),
		sizeof (epicsFloat32),
		sizeof (epicsFloat64),
		sizeof (epicsString),
		sizeof (epicsOldString),
};
#else /* epicsTypesGLOBAL */
epicsShareExtern READONLY unsigned epicsTypeSizes [lastEpicsType+1];
#endif /* epicsTypesGLOBAL */

/*
 * The enumeration "epicsType" is an index to this array
 * of type class identifiers.
 */
typedef enum {
	epicsIntC, 
	epicsUIntC, 
	epicsEnumC,
	epicsFloatC, 
	epicsStringC,
	epicsOldStringC} epicsTypeClass;
#ifdef epicsTypesGLOBAL
READONLY epicsTypeClass epicsTypeClasses [lastEpicsType+1] = {
		epicsIntC,
		epicsUIntC,
		epicsIntC,
		epicsUIntC,
		epicsEnumC,
		epicsIntC,
		epicsUIntC,
		epicsFloatC,
		epicsFloatC,
		epicsStringC,
		epicsOldStringC
	};
#else /* epicsTypesGLOBAL */
epicsShareExtern READONLY epicsTypeClass epicsTypeClasses [lastEpicsType+1];
#endif /* epicsTypesGLOBAL */


#ifdef epicsTypesGLOBAL
READONLY char *epicsTypeAnyFieldName [lastEpicsType+1] = {
			stringOf (int8),
			stringOf (uInt8),
			stringOf (int16),
			stringOf (uInt16),
			stringOf (enum16),
			stringOf (int32),
			stringOf (uInt32),
			stringOf (float32),
			stringOf (float64),
			stringOf (string),
			"", /* Old Style Strings will not be in epicsAny type */
	};
#else /* epicsTypesGLOBAL */
epicsShareExtern READONLY char *epicsTypeAnyFieldName [lastEpicsType+1];
#endif /* epicsTypesGLOBAL */

#endif /* INCepicsTypesh */

